/*
 * Copyright (c) 2005 The University of Reading
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University of Reading, nor the names of the
 *    authors or contributors may be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package uk.ac.rdg.resc.jstyx.server;

import net.tinyos.util.*;
import net.tinyos.message.*;
import net.tinyos.tools.*;
import java.util.*;
import java.lang.Thread;

import org.apache.log4j.Logger;

import org.apache.mina.common.ByteBuffer;

import uk.ac.rdg.resc.jstyx.StyxUtils;
import uk.ac.rdg.resc.jstyx.StyxException;
import uk.ac.rdg.resc.jstyx.types.ULong;

/**
 * File whose underlying data are stored as a ByteBuffer in memory. This buffer
 * can grow to arbitrary size.
 *
 * @author Jon Blower
 * $Revision: 1.19 $
 * $Date: 2005/09/08 07:08:59 $
 * $Log: InMemoryFile.java,v $
 * Revision 1.19  2005/09/08 07:08:59  jonblower
 * Removed "String user" from list of parameters to StyxFile.write()
 *
 * Revision 1.18  2005/09/02 16:51:20  jonblower
 * Fixed bug with ByteBuffers being released prematurely and changed to use autoExpanding ByteBuffers
 *
 * Revision 1.17  2005/09/01 07:50:22  jonblower
 * Trying to fix bug with ByteBuffers being released (can't read from InMemoryFile)
 *
 * Revision 1.16  2005/08/30 16:29:00  jonblower
 * Added processAndReplyRead() helper functions to StyxFile
 *
 * Revision 1.13  2005/06/10 07:53:12  jonblower
 * Changed SGS namespace: removed "inurl" and subsumed functionality into "stdin"
 *
 * Revision 1.12  2005/04/28 08:11:15  jonblower
 * Modified permissions handling in documentation directory of SGS
 *
 * Revision 1.10  2005/04/27 16:11:43  jonblower
 * Added capability to add documentation files to SGS namespace
 *
 * Revision 1.9  2005/04/26 07:46:11  jonblower
 * Continuing to improve setting of parameters in Styx Grid Services
 *
 * Revision 1.8  2005/03/24 15:11:07  jonblower
 * Changed so that underlying ByteBuffer is allocated on the first write to the file
 *
 * Revision 1.7  2005/03/24 13:07:09  jonblower
 * Added code to prevent the file growing to larger than a specified size
 *
 * Revision 1.6  2005/03/24 12:55:14  jonblower
 * Changed to use ByteBuffer as backing store
 *
 * Revision 1.5  2005/03/24 09:48:31  jonblower
 * Changed 'count' from long to int throughout for reading and writing
 *
 * Revision 1.4  2005/03/16 17:56:23  jonblower
 * Replaced use of java.nio.ByteBuffer with MINA's ByteBuffer to minimise copying of buffers
 *
 * Revision 1.3  2005/03/11 14:02:16  jonblower
 * Merged MINA-Test_20059309 into main line of development
 *
 * Revision 1.2.2.1  2005/03/10 20:55:38  jonblower
 * Removed references to Netty
 *
 * Revision 1.2  2005/03/01 13:47:43  jonblower
 * Changed default user and group to 'user' and 'group'
 *
 * Revision 1.1.1.1  2005/02/16 18:58:31  jonblower
 * Initial import
 *
 */
public class ClusterFS 
{    

    //   public static final short TOS_BCAST_ADDR = (short) 0xffff;
    public static final short TOS_BCAST_ADDR = (short) 0;
    public static final byte S_Tread = (byte) 0;
    public static final byte S_Twrite = (byte) 2;
    public static final byte S_Rwrite = (byte) 3;
    public static final byte S_Revent = (byte) 4;
    public static final byte S_Rerror = (byte) 5;
    public static final byte S_Rnoid  = (byte) 6;

    public static final byte INFO_RID = (byte) 0;
    public static final int NO_RSP = 0;
    public static final byte TEMP_RID = (byte) 1;
    public static final byte PHOTO_RID = (byte) 2;
    public static final byte TEVNT_RID = (byte) 3;
    public static final byte PEVNT_RID = (byte) 4;

    // needs to be consistent with the above rid's
    public static String[] RIDNAMES = {"tempsensor", "photosensor", "tempevent", "photoevent"};

    /* maps the files above to their sensor types, surely there are better ways of doing this ...*/
    public static int[] RIDTYPES = { 0, 1, 0, 1};

    public static EPMsgNotifier rtnNotifier;
    //    private static EPMote[] mote = {new EPMote(0), new EPMote(1), new EPMote(2), new EPMote(3)};
    private static EPMote[] mote;
    ///    private static EPMote[] mote;

    /*
    public static EPMote getMote(int index) {
	return mote[index];
    }
    */

    public static boolean isEventResource(int rid)
    {
	return ((rid == (int)TEVNT_RID) || (rid == (int)PEVNT_RID));
    }

    public static void createNS(StyxDirectory dir, int sensorno, EPMote epm, StyxDirectory[] senTypesDir, Vector AggrFileVec)
    {
	/* create a new packet to send off ... */
	byte buffer[];
	int size=10;
	buffer=new byte[size];
	SensorFile discoveredFile;
	int rtnarg;

	EPMsg outpacket = new EPMsg(buffer); 
	int ridindex = 0;

	EPMsgR returnpacket = new EPMsgR();
      
	System.out.println("createNS 1.1\n");
		
	while (true) 
         {
		outpacket.set_arg((byte)ridindex);  //incremented by 1 each time
		outpacket.set_type(S_Tread); //define operation to be 'read'
		outpacket.set_rid(INFO_RID);  
		//		inpacket.set_sensorno((byte)0);  //target sensor #0
			
		System.err.println("Sending payload: "+((EPMsg)outpacket).toString());
		System.err.println("To sensor: " + sensorno);             				    
			
		try {
		    mote[sensorno].sendrecv(outpacket, returnpacket);
		    rtnarg = returnpacket.get_arg();
		    if (rtnarg == NO_RSP)
			break;
		    else {
			if (isEventResource(rtnarg)) {
			    System.out.println("discovered Eventfile");
			    discoveredFile = new EventFile(RIDNAMES[rtnarg - 1] + java.lang.String.valueOf(sensorno), sensorno, epm, rtnarg);
			}
			else {
			    System.out.println("discovered Sensorfile");
			    discoveredFile = new SensorFile(RIDNAMES[rtnarg - 1] + java.lang.String.valueOf(sensorno), sensorno, epm, rtnarg);
			}
			dir.addChild(discoveredFile);

			/* add this sensor to the approp group */
			senTypesDir[RIDTYPES[rtnarg - 1]].addChild(discoveredFile);

			/* update the aggr file lists */
			Enumeration e = AggrFileVec.elements();
			while (e.hasMoreElements()) {
			    AggrFile aggrfile = (AggrFile) e.nextElement();
			    
			    if (rtnarg  == aggrfile.rid)
				aggrfile.addSensorFile(discoveredFile);
			}
			
			//		System.out.println("file :"+ RIDNAMES[rtnarg - 1]);
			ridindex++;
		    }
		}
		catch(Exception e) 
		    {
			e.printStackTrace();
		    }	
	 }
		
		
    }
	
    
    public static void main (String[] args) throws Exception
    {
        // Create the root directory of the Styx server
        StyxDirectory root = new StyxDirectory("/");

	/* additionally each sensor has its own directory       */
	StyxDirectory[] sensorDirArray;

	/* each sensor type has its own directory        */
	StyxDirectory[] sensorTypeDirArray;
	StyxDirectory sensor0dir = new StyxDirectory("sensor0");

	/* create all aggregate files here: for now avg temperature  */
	AvgAggrFile tempAvgFile = new AvgAggrFile("tempaverage", TEMP_RID);
	root.addChild(tempAvgFile);
	/* create a vector to store all these aggregate files together */
	Vector AggrFileVec;
	AggrFileVec = new Vector();
	AggrFileVec.add(tempAvgFile);
	
	/* object that notifies all virtual motes once a return message is received   */
	rtnNotifier = new EPMsgNotifier();
        sensorDirArray = new StyxDirectory[Integer.parseInt(args[0])];


	/* directory for grouping all like sensors together  */
	StyxDirectory tempDir;
	StyxDirectory photoDir;

	tempDir = new  StyxDirectory("tempSensors");
	photoDir = new StyxDirectory("photoSensors");

	/* we have 2 types of sensors, maybe automate this later ...  */
	sensorTypeDirArray = new StyxDirectory[2];
	sensorTypeDirArray[0] = tempDir;
	sensorTypeDirArray[1] = photoDir;
	root.addChild(tempDir);
	root.addChild(photoDir);


	//	mote = new EPMote[4];
	mote = new EPMote[Integer.parseInt(args[0])];

	if (args.length != 1) 
	    {
		System.out.println("Improper usage: <prog name> <number of motes>");
		//System.out.println("you input "+argv.length+" number of args = "+argv[0]);
		System.exit(-1);
	    }

	for(int moteindex = 0; moteindex < Integer.parseInt(args[0]); moteindex++) 
	    {
		sensorDirArray[moteindex] = new StyxDirectory("sensor" + String.valueOf(moteindex));
		
		mote[moteindex] = new EPMote(moteindex, rtnNotifier);
		
		createNS(sensorDirArray[moteindex], moteindex, mote[moteindex], sensorTypeDirArray, AggrFileVec);
		
		// Add a sensor file to the root
		root.addChild(sensorDirArray[moteindex]);
	    }	
		// Start a StyxServer, listening on port 9876
		new StyxServer(9876, root).start();

    }
    
}
