/*
    FUSE: Filesystem in Userspace
    Copyright (C) 2001-2005  Miklos Szeredi <miklos@szeredi.hu>

    This program can be distributed under the terms of the GNU GPL.
    See the file COPYING.
*/

#define FUSE_USE_VERSION 25
#include <fuse.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include "slib.h"
#include "mystyx.h"

static int rootfid;
static Qid rootqid;
int styx_rd, styx_wr;

/*
  A structure to store file descriptors of local file, so that we don't open them over and over 
     again.
*/
struct fid_table_entry {
  int fd;
  char path[DIRLEN];
  Dir d;
};

typedef struct fid_table_entry fid_table_entry_t;

/*
 size of this table obviously depends on the number of files in the imported filesystem,
    for now set it to be 20
*/
fid_table_entry_t fid_table[20]; 

/*
 initialize this to a state where there are no fids
*/
static init_fid_table()
{
  int i;

  for(i = 0; i < 20; i++)
    fid_table[i].fd = -1;

}

/*
  given a path return the already opened fd that it might have
*/
fid_table_entry_t *find_fid_tab_entry(const char *path)
{
  int i;

  for(i = 0; i < 20; i++)
    if (!strcmp(path, fid_table[i].path))
      return &(fid_table[i]);
  return NULL;
}



int get_stat(int fid, Dir *dirptr)
{
  int i;

  for(i = 0; i < 20; i++)
    if (fid_table[i].fd == fid) {
      memcpy(dirptr, &fid_table[i].d, sizeof(Dir));
      return STYX_OK;
    }
  return STYX_ERROR;
}


void make_fid_table_entry(int pathfid, const char *path, Dir *dirptr)
{
  int i;

  for(i = 0; i < 20; i++)
    if (fid_table[i].fd == -1) {
      strcpy(fid_table[i].path, path);
      fid_table[i].fd = pathfid;
      memcpy(&fid_table[i].d, dirptr, sizeof(Dir));
      return;
    }
  printf("Warning: unable to add fid\n");
}

/*
  change this so that each remote file is walked to only once and after that reused for all future 
     opens
*/

static int pathtofid(const char *path)
{
  int pathfid;
  char *p;
  char *next;
  char *cp;
  int res;
  fid_table_entry_t *fte;
  char *pathorig;
  Dir st;

  pathorig = path;
  fte = find_fid_tab_entry(path);

  /*
    if fid already exists just return it
  */
  if (fte)
    return (fte->fd);

  pathfid = getFid();

  if ((pathfid < 0) || (path == 0) || (path[0] == 0))
    return STYX_ERROR;
  
  if ((res = styx_clone(rootfid, pathfid)) == STYX_ERROR)
    {
      releaseFid(pathfid);
      return STYX_ERROR;
    }

  if (p = strdup(path))
    {
      Qid wqid;
      if (p[0] == '/') path++;
      for (next = strtok_r(p, "/",  &cp); next; next = strtok_r(0, "/", &cp))
	{
	  if ((res = styx_walk(pathfid, next, &wqid)) == STYX_ERROR)
	    break;
	}
    }
  else 
    res = STYX_ERROR;

  if (p) free(p);
  if (res == STYX_ERROR)
    {
      styx_clunk(pathfid);
      releaseFid(pathfid);
      return STYX_ERROR;
    }

  /*
    if we can get the stat, then make entry in fid table
  */

  if (styx_stat(pathfid, &st) == STYX_OK) {
    make_fid_table_entry(pathfid, pathorig, &st);
    return pathfid;
  }

  /* if we got here then something bad happened */
  styx_clunk(pathfid);
  releaseFid(pathfid);
  return STYX_ERROR;

}


static int sf_getattr(const char *path, struct stat *stbuf)
{
    int res = 0;
    int tmpfid = pathtofid(path);
    if ((tmpfid = pathtofid(path)) < 0)
      {
	res = -ENOENT;
      }
    else
      {
	Dir st;
	memset(stbuf, 0, sizeof(struct stat));
	if (get_stat(tmpfid, &st) == STYX_OK)
	  {
	    if (st.mode & DMDIR)
	      stbuf->st_mode = S_IFDIR;
	    else
	      stbuf->st_mode = S_IFREG;
	    stbuf->st_mode |= (st.mode & 0777);
	    stbuf->st_nlink = 1;              // this may need some work
	    stbuf->st_size = st.length;
	    stbuf->st_atime = st.atime;
	    stbuf->st_mtime = st.mtime;
	  }
	else
	  res = -ENOENT;
	//	styx_clunk(tmpfid);
	releaseFid(tmpfid);
      }
    return res;
}



static int sf_open(const char *path, struct fuse_file_info *fi)
{
  int fid = pathtofid(path);
  if (fid >= 0)
    {
      fi->fh = fid;
      if (styx_open(fid, fi->flags&0x3) >= 0)
	return 0;
      else
	{
	  //	  styx_clunk(fid);
	  releaseFid(fid);
	  return 0;
	}
    }
  else
    return -ENOENT;
}


static int sf_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
                         off_t offset, struct fuse_file_info *fi)
{
    (void) offset;
    (void) fi;

    if (sf_open(path, fi) >= 0)
      {
	Dir stat;
	char sbuf[sizeof(Dir)];
	int i;
        filler(buf, ".", NULL, 0);
	filler(buf, "..", NULL, 0);

	for (i = 0; 
	     styx_read((int)(fi->fh), i, sizeof(Dir), sbuf);
	     i = i + sizeof(Dir)
	     )
	  {
	    convM2D(sbuf, &stat);
	    filler(buf, stat.name, NULL, 0);
	  }
	//	styx_clunk(fi->fh);
	releaseFid(fi->fh);
	fi->fh = 0;
	return 0;
      }

    return -ENOENT;
}


static int sf_read(const char *path, char *buf, size_t size, off_t offset,
                      struct fuse_file_info *fi)
{
  if (fi->fh)
    {
      return styx_read(fi->fh, offset, size, buf);
    }
  else
    return -ENOENT;
}

static int sf_write(const char *path, const char *buf, size_t size,
                     off_t offset, struct fuse_file_info *fi)
{
  if (fi->fh)
    {
      return styx_write(fi->fh, offset, size, buf);
    }
  else
    return -ENOENT;
}

static int cnct(char *servername, int portno)
{
  struct sockaddr_in serv_addr;
  struct hostent* server;
  int sockfd;

  sockfd = socket(AF_INET, SOCK_STREAM, 0);
  server = gethostbyname(servername);
  bzero((char*) &serv_addr, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  bcopy((char*) server->h_addr,
        (char*) &serv_addr.sin_addr.s_addr,
        server->h_length);
  serv_addr.sin_port = htons(portno);

  // ready to connect now

  if(connect(sockfd, (struct sockaddr*)&serv_addr, sizeof(serv_addr)) < 0) {
    fprintf(stderr, "ERROR connecting\n");
    exit(0);
  }
  styx_rd = sockfd;
  styx_wr = sockfd;
  fprintf(stderr,"sockfd %d\n", sockfd);
}


static struct fuse_operations sf_oper = {
  .getattr= sf_getattr,
  .readdir= sf_readdir,
  .open= sf_open,
  .read= sf_read,
  .write = sf_write
};


main(int argc, char** argv)
{


  if (styx_connect_serial() == STYX_ERROR) {
    printf("Problems opening serial port, exiting\n");
    exit(-1);
  }
  printf("opened serial port\n");
  init_fid_table();   
  rootfid = getFid();

  if (styx_attach(rootfid, &rootqid) == STYX_ERROR) {
    fprintf(stderr,"unable to attach, exiting.\n");
    exit(-1);
  }

  if (styx_clone(rootfid, 4) == STYX_ERROR)
    {
      printf("trouble with initial clone\n");
    }

  printf("attached to filesystem\n");
  return fuse_main(argc, argv, &sf_oper);
}
