#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <assert.h>
#include <signal.h>
#include "gdb.h"
#include "arm7target.h"
#include "arm7tdmi.h"

/* max read size  */
#define MAXMEM 400

/* bounds for RAM*/
#define RAMSTART 0x40000000
#define RAMEND 0x40003FFF

/* escape chars for binary transfer  */
#define ESCAPE_BYTE 0x7d
#define ESCAPE_XOR_MASK 0x20

/* number of breakpoints */
#define NUM_BREAKPOINTS 2

extern register_file_S gdb_register_file;

const char *register_names[] = {"R0", "R1", "R2", "R3", "R4", "R5", "R6", "R7", \
                         "R8", "R9", "R10", "R11", "R12", "R13", "R14", "PC"};

const char FUSE_MOUNT_PATH[] = "/home/bpisupat/mnt/armdevdir";

long *register_file_ptrs[15];

char lnibble_to_hex (char i);
int hex_to_int (char h);
void last_signal (int sigval);

/*  
    store which breakpoints are free
*/
unsigned char bkpt_array[NUM_BREAKPOINTS];

/*
  File descriptors for the various files that we will open in the arm file system
*/
int registersfid;
int memoryfid;
int statusfid;
int controlfid;



int ishalted() 
{
  char status;
  int retval;

  lseek(statusfid, 0, SEEK_SET);
  retval = read(statusfid, &status, 1);
  if (retval < 1) {
    printf("warning: ishalted is unable to get status of target\n");
    return 0;
  }
  if (status == 's')
    return 1;
  else return 0;
}

/*
  TODO: the lcbuf unions assume and depend that lbuf and sbuf start at
  the same address.  Is this always correct?  Is there a better way?
*/

/* Converts a memory region of length len bytes, starting at mem, into
   a string of hex bytes.  Returns the number of bytes placed into
   hexbuf.

   This function carefully preserves the endianness of the data,
   because that's what gdb expects.  This function also optimizes the
   read process into the largest units possible, in case we're reading
   a peripheral register that can't deal with unaligned or byte-wide
   accesses.  */
int mem_to_hexbuf (const void* mem, char* hbuf, int len) {
  int retval = 0;
  while (len > 0) {
    /* byte 0 - LSB */
    *hbuf++ = lnibble_to_hex(*((unsigned char *) mem) >> 4);
    *hbuf++ = lnibble_to_hex(*((unsigned char *) mem++));

    len--;
    retval += 2;
  }
  
  return retval;
}



/*
  Reads (len * 2) hex digits from hbuf, converts them to binary,
  writes them to mem. Returns a pointer to the first empty byte after
  the region written.

  Carefully preserves endianness, optimizes write accesses so as to be
  hardware-friendly.

*/
char* hexbuf_to_mem (char* hbuff, void* mem, int len) {
  while (len > 0) {
    /* byte 0 - LSB */
    *((unsigned char *)mem)++ = (hex_to_int(*hbuff++) << 4) + hex_to_int(*hbuff++);
    len--;
  }
  return hbuff;
}

/* just copy one byte after another until an escape char is received
     in which case XOR the next byte with 0x20  */
char* binbuf_to_mem(unsigned char* binbuf, void* membuf, int len) {
  unsigned char thisbyte;

  while (len > 0) {
    if ((thisbyte = *binbuf++) == ESCAPE_BYTE)
      *((unsigned char *)membuf)++ = (*binbuf++) ^ ESCAPE_XOR_MASK;
    else
      *((unsigned char *)membuf)++ = thisbyte;
    len--;
  }
  return binbuf;
}

void set_breakpoint(u32 addr) {
  char command;
  char breakwrite[12];
  int retval;

  breakwrite[0] = 'b'; /* the break command */
  breakwrite[1] = '0'; /* bkpoint number - for now irrelevant */
  breakwrite[2] = lnibble_to_hex((addr & 0xF0000000) >> 28);
  breakwrite[3] = lnibble_to_hex((addr & 0xF000000) >> 24);
  breakwrite[4] = lnibble_to_hex((addr & 0xF00000) >> 20);
  breakwrite[5] = lnibble_to_hex((addr & 0xF0000) >> 16);
  breakwrite[6] = lnibble_to_hex((addr & 0xF000) >> 12);
  breakwrite[7] = lnibble_to_hex((addr & 0xF00) >> 8);
  breakwrite[8] = lnibble_to_hex((addr & 0xF0) >> 4);
  breakwrite[9] = lnibble_to_hex(addr & 0xF);

  retval = write(controlfid, breakwrite, 10);
  if (retval < 10) 
    printf("Warning: unable to set breakpoint at %x\n", addr);

} 
  
  
void remove_breakpoint(u32 addr) {

  char command;
  char breakwrite[12];
  int retval;

  breakwrite[0] = 'b'; /* the break command */
  breakwrite[1] = '0'; /* bkpoint number - for now irrelevant */
  breakwrite[2] = '0';
  breakwrite[3] = '0';
  breakwrite[4] = '0';
  breakwrite[5] = '0';
  breakwrite[6] = '0';
  breakwrite[7] = '0';
  breakwrite[8] = '0';
  breakwrite[9] = '0';

  retval = write(controlfid, breakwrite, 10);
  if (retval < 10) 
    printf("Warning: unable to set breakpoint at %x\n", addr);


}
 
/*
  Converts the escaped-binary ('X' packet) array pointed to by buf
  into binary, to be placed in mem. Returns a pointer to the first
  empty byte after the region written.
*/
char* xmem_to_mem (const char* xmem, void* mem, int len)	{
  return 0;
}

void write_memory (const char *hexbuf, long addr, long len) {
 unsigned char membuf[MAXMEM];
 int retval;


 if ((addr > RAMEND) || (addr < RAMSTART)) {
   printf("WARNING: writing to illegal address %p, ignoring write\n", addr);
   gdbputmsg(0, "OK", 2);
   return;
 }

 /* set offset to write address */
 lseek(memoryfid, addr, SEEK_SET);
 hexbuf = hexbuf_to_mem(hexbuf, membuf, len);
 retval = write(memoryfid, membuf, len);
 if (retval < len) 
   printf("Warning: trouble writing memory, tried to write %d bytes, wrote only %d\n", len, retval);
 
 
 gdbputmsg(0, "OK", 2);
 return;
 
}

/* accepts an escaped binary buffer in binbuf     */
void write_xbin_memory (const char *binbuf, long addr, long len) {
 unsigned char membuf[MAXMEM];
 //usigned char readmembuf[MAXMEM];
 u32 readmem;
 int retval;

 int num32, num16, num8;
 int i;

 if ((addr > RAMEND) || (addr < RAMSTART)) {
   printf("WARNING: writing to illegal address %p, ignoring write\n", addr);
   gdbputmsg(0, "OK", 2);
   return;
 }

 /* set offset to write address */
 lseek(memoryfid, addr, SEEK_SET);
 binbuf = binbuf_to_mem((unsigned char *) binbuf, membuf, len);
 retval = write(memoryfid, membuf, len);
 if (retval < len) 
   printf("Warning: trouble writing memory, tried to write %d bytes, wrote only %d\n", len, retval);
 
   
  gdbputmsg(0, "OK", 2);
  return;

}

void read_memory (long addr, long len) {
  char tx_buf[MAXMEM * 2];
  int tx;
  unsigned char sum = 0;
  unsigned char membuf[MAXMEM];
  int retval;
  
  /* parse address, length */


  gdb_putc('$');
  sum = 0; 
 
  if ((addr > RAMEND) || (addr < RAMSTART)) {
    printf("WARNING: reading from illegal address %p\n", addr);
    bzero(membuf, len);
    tx = mem_to_hexbuf(membuf, tx_buf, len);
    sum += putbuf(tx, tx_buf);
    gdb_putc('#');
    gdb_putc(lnibble_to_hex(sum >> 4));
    gdb_putc(lnibble_to_hex(sum));

    while (gdb_getc() != '+');
    return;
  }

  lseek(memoryfid, addr, SEEK_SET);
  retval = read(memoryfid, membuf, len);
  if (retval < len) 
    printf("Warning: trouble writing memory, tried to write %d bytes, wrote only %d\n", len, retval);
  tx = mem_to_hexbuf((u8 *)membuf, tx_buf, retval);
  sum += putbuf(tx, tx_buf);
    
  gdb_putc('#');
  gdb_putc(lnibble_to_hex(sum >> 4));
  gdb_putc(lnibble_to_hex(sum));

  while (gdb_getc() != '+')
    usleep(50);
  return;
}

void gdb_singlestep() {
  char command;
  int retval;

  command = 't'; /* the break command */

  printf("single step:: restarting from %x\n", *register_file_ptrs[15]);
  write_registers_totarget();

  retval = write(controlfid, &command, 1);
  if (retval < 1) 
    printf("Warning: unable to single step\n");

   while (!(ishalted()))
    usleep(100);
  read_registers_fromtarget();
  // putmsg(0, "OK", 2);
  //last_signal(SIGSTOP);
  last_signal(SIGTRAP);
  return;
}


/* probably called after break, single step and stop  */
int read_registers_fromtarget() {
  int i;
  int retval = 0;
  char registers_val[8*16]; /* 15 registers, each 8 bytes in size */


  for(i = r0; i<= r15; i++) {
    lseek(registersfid, i, SEEK_SET);
    retval = read(registersfid, registers_val, 8);
    if (retval < 8) 
      printf("Warning: trouble writing memory, tried to write 120 bytes, wrote only %d\n", retval);
    

    *register_file_ptrs[i] = (long) ((hex_to_int(registers_val[0]) << 28) + (hex_to_int(registers_val[1]) << 24) + (hex_to_int(registers_val[2]) << 20) + (hex_to_int(registers_val[3]) << 16) + (hex_to_int(registers_val[4]) << 12) + (hex_to_int(registers_val[5]) << 8) + (hex_to_int(registers_val[6]) << 4) + (hex_to_int(registers_val[7])));
  }

}

/* probably called just before, continue or single step */
int write_registers_totarget() {
  int i;
  int retval = 0;
  char registers_val[8*16]; /* 15 registers, each 8 bytes in size */

  for(i = r0; i<= r15; i++) {
    lseek(registersfid, i, SEEK_SET);
    registers_val[0] = lnibble_to_hex((*register_file_ptrs[i] & 0xF0000000) >> 28);
    registers_val[1] = lnibble_to_hex((*register_file_ptrs[i] & 0xF000000) >> 24);
    registers_val[2] = lnibble_to_hex((*register_file_ptrs[i] & 0xF00000) >> 20);
    registers_val[3] = lnibble_to_hex((*register_file_ptrs[i] & 0xF0000) >> 16);
    registers_val[4] = lnibble_to_hex((*register_file_ptrs[i] & 0xF000) >> 12);
    registers_val[5] = lnibble_to_hex((*register_file_ptrs[i] & 0xF00) >> 8);
    registers_val[6] = lnibble_to_hex((*register_file_ptrs[i] & 0xF0) >> 4);
    registers_val[7] = lnibble_to_hex(*register_file_ptrs[i] & 0xF);
    retval = write(registersfid, registers_val, 8);
    if (retval < 8)
      printf("warning : write_register to reg %d, tried to write 8, wrote %d\n", i, retval);
    
  }


}


void gdb_handle_exception (int sigval) {}

void gdb_continue (long addr) 
{
  char command = 'c';
  int retval;

  printf(" continue :: restarting from %x\n", addr);
  write_registers_totarget();

  retval = write(controlfid, &command, 1);
  if (retval < 1) 
    printf("Warning: unable to single step\n");

  while (!(ishalted()))
    usleep(100);
  read_registers_fromtarget();
  //  last_signal(SIGSTOP);
  last_signal(SIGTRAP);
  
  return;
}

void gdb_kill (void){}
void gdb_detach (void){}
void gdb_return_from_exception (void) {}
void gdb_flush_cache (void *start, int len){}

void target_initialize() {

  char path[40];

  /* for now all initialization happens automatically within the filesystem, so just open the files */
  sprintf(path, "%s/registers", FUSE_MOUNT_PATH);
  registersfid = open(path, O_RDWR);

  sprintf(path, "%s/memory", FUSE_MOUNT_PATH);
  memoryfid = open(path, O_RDWR);

  sprintf(path, "%s/status", FUSE_MOUNT_PATH);
  statusfid = open(path, O_RDWR);

  sprintf(path, "%s/control", FUSE_MOUNT_PATH);
  controlfid = open(path, O_RDWR);

   /* the way gdb_register_file is structured necessitates this rather dumb way
       of initializing...  */
   register_file_ptrs[0] = &gdb_register_file.r0;
   register_file_ptrs[1] = &gdb_register_file.r1;
   register_file_ptrs[2] = &gdb_register_file.r2;
   register_file_ptrs[3] = &gdb_register_file.r3;
   register_file_ptrs[4] = &gdb_register_file.r4;
   register_file_ptrs[5] = &gdb_register_file.r5;
   register_file_ptrs[6] = &gdb_register_file.r6;
   register_file_ptrs[7] = &gdb_register_file.r7;
   register_file_ptrs[8] = &gdb_register_file.r8;
   register_file_ptrs[9] = &gdb_register_file.r9;
   register_file_ptrs[10] = &gdb_register_file.r10;
   register_file_ptrs[11] = &gdb_register_file.r11;
   register_file_ptrs[12] = &gdb_register_file.r12;
   register_file_ptrs[13] = &gdb_register_file.r13;
   register_file_ptrs[14] = &gdb_register_file.r14;
   register_file_ptrs[15] = &gdb_register_file.r15;
}
