/*
 * main.c : the main source file for JTAGER
 *
 * Copyright (C) 2003, Rongkai Zhan <zhanrk@163.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: main.c,v 1.1.1.1 2007/06/05 16:01:43 bpisupat Exp $ */

/*
 * References:
 * [1] IEEE Std 1149.1-1990, <<IEEE Standard Test Access Port and 
 *      Boundary-Scan Architecture>>
 * [2] IEEE Std 1149.1b-1994, <<Supplement to IEEE Std 1149.1-1990>>
 * [3] <<ARM Architecture Reference Manual>>
 * [4] <<ARM7TDMI Technical Reference Manual>> Rev 4
 * [5] <<ARM9TDMI Technical Reference Manual>> Rev 3
 * [6] Application Note 28, <<The ARM7TDMI Debug Architecture>> 
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "jtag.h"
#include "target.h"
#include "arm7tdmi.h"

/*
 * The character buffer for jtager command line input...
 * It is terminated by NULL character '\0', therefore we get 2
 */

extern int do_reg_read_all();
extern int halt_run(int argc, char **argv);
extern int checkhalt();

const u32 opcode_idelay[] = {
  0xe92d4030,
  0xe3a0320e,
  0xe283390a,
  0xe5932038,
  0xe24dd004,
  0xe3822001,
  0xe3a01000,
  0xe5832038,
  0xe58d1000,
  0xe1a05003,
  0xe3a04001,
  0xe5854034,
  0xebffffe9,
  0xe585403c,
  0xebffffe7,
  0xe59d3000,
  0xe2833001,
  0xe58d3000,
  0xeafffff7,
  NULL
};

static char *err_desc[] = {
	"\tunknown error.\n", /* ERR_COMMON */

	"\tInvalid function parameters.\n", /* ERR_INVALID_PARAM */

	/* ERR_JTAG_INSTRUCTION_FAILED */
	"\tFailed to read/write the JTAG instruction register. The JTAG"\
	"\n\tinstruction register always outputs a fixed known value during"\
	"\n\tthe CAPTURE-DR stage. E.g. for ARM7TDMI target, its JTAG"\
	"\n\tinstruction register always outputs b0001. But now we read a"
	"\n\tdifferent value. Something is wrong.\n",

	/* ERR_JTAG_INVALID_IDCODE */
	"\tWe read an invalid idcode value from the JTAG IDCODE register."\
	"\n\tThe bit[0] of idcode value must be 1.\n",

	/* ERR_JTAG_NONEXISTED_SCANPATH */
	"\tYour specified scan chain is non-existed on your target.\n",

	/* ERR_JTAG_SCANPATH_FAILED */
	"\tFailed to select your specified scan chain. The JTAG scan path"\
	"\n\tselect register always outputs a fixed known value during the"\
	"\n\tCAPTURE-DR stage. E.g. for ARM7TDMI target, its JTAG scan path"\
	"\n\tselect register always outputs b1000. But now we read a"\
	"\n\tdifferent value. Something is wrong.\n",

	/* ERR_TARGET_NONEXISTED_ICE_REG */
	"\tYour specified ICE register is non-existed on your target. Maybe"\
	"\n\tyou have specified an illegal ICE register address.\n",

	/* ERR_TARGET_HALT_FAILED */
	"\tFailed to halt the target and make it into the debug state.\n",

	/* ERR_TARGET_IS_RUNNING */
	"\tThe target is running. Please use \"halt\" command to halt it.\n",

	/* ERR_TARGET_IN_THUMB_MODE */
	"\tThe target is in THUMB mode, which is not implemented by JTAGER.\n",

	/* ERR_TARGET_UNKNOWN_REG */
	"\tUnknown core register name.\n",

	/* ERR_TARGET_UNKNOWN_ICE_REG */
	"\tUnknown ICE register name.\n",

	/* ERR_TARGET_INVALID_CPSR */
	"The CPSR register's value is undetermined and invalid. Before we"\
	"\n\tcan read or write SPSR register, we must have have an valid CPSR"\
	"\n\tregister value. And you can use command \"reg cpsr\" to get it.\n",

	/* ERR_TARGET_NO_SPSR */
	"\tThe target is in USER mode, so it has no SPSR register.\n",

	/* ERR_FLASH_TIMEOUT */
	"\tTime out.\n",

	NULL /* the last one must be NULL, to terminate the array */
};

char *err_str(int retval)
{
	int index = -retval - 1;
	
	return err_desc[index];
}




void armmain1() 
{
  u32 idcode;
  int ret = 0;
  core_register_t *reg = NULL;
  u32 memval = 0x67452301;
  u32 memvalret;
  u32 memread;
  u32 regval;
 
  /* Initialize jtag */
  if (jtag_reset()) {
    printf("Error: %s: failure to initialize jtag ...\n");
    exit(-1);
  }

  arm7tdmi_setup();
  if (jtag_idcode(&idcode)) {
    printf("error in obtaining IDCODE, exiting\n");
    return;
  }
  printf("device idcode = %x\n",idcode);
  ret = arm7tdmi_halt();
  if (ret)
    printf("halt: %s", err_str(ret)); 
  printf("halted successfully\n");
 
 
  reg = reg_index("PC");
  ret = arm7tdmi_register_read(reg);
  if (ret) {
    //  printf("Error:%s", err_str(ret));
    printf("Error reading PC\n");
  } else {
    printf("reading PC = %x\n", reg->value);
  } 
 /*
  regval = 0x12345678;
  reg = reg_index("R1");
 
  ret = arm7tdmi_register_write(reg, regval);
  ret = arm7tdmi_register_read(reg);
*/

  ret = arm7tdmi_memory_write32(&opcode_idelay, 0x40000308, 19); // idelay
  ret = arm7tdmi_memory_read32(&memread, 0x40000310, 1); // idelay
  
  if (arm7tdmi_ice_set_addrbreakpt(0x40000334))
     printf("trouble setting breakpoint at addr\n");
     
  arm7tdmi_restart(0x40000308); //once + idelay
  
  if (checkhalt())
    printf("checkhalt failed\n");
  else printf("halted\n");
 
 /* 
  ret = arm7tdmi_register_read(reg);
  if (ret) {
    //  printf("Error:%s", err_str(ret));
    printf("Error reading PC\n");
  } else {
    printf("after bk reading PC = 0x%.8lX\n", reg->value);
  }
  */ 
  arm7tdmi_set_ss();
  
 // ret = arm7tdmi_memory_read32(&memread, 0x40000300, 1); // idelay
  
 // printf("read val = %x\n", reg->value);
 

}

  
