/*========================================================================*\
|                                                                          |
| cycles.c                                                                 |
|                                                                          |
| This file contains functions that determine the number of CPU cycles     |
| required to execute a specified instruction.                             |
|--------------------------------------------------------------------------|
| Project:              MSP430 JTAG interfcae                              |
| Developed using:      MS Visual C++ 5.0                                  |
|--------------------------------------------------------------------------|
| Author:               UPSF                                               |
| Version:              1.01                                               |
| Initial Version:      20 / 07 / 02                                       |
| Last Change:                                                             |
|--------------------------------------------------------------------------|
| Version history:                                                         |
| Version: 1.01                                                            |
| 10.02.2003 UPSF fixed counts for BR #xxxx                                |
|                                                                          |
|--------------------------------------------------------------------------|
| Designed 2002 by Texas Instruments                                       |
\*========================================================================*/

/*------------------------------------------------------------------------*\
| Remarks:                                                                 |
|                                                                          |
\*------------------------------------------------------------------------*/

// #includes. -----------------------------------------------------------------

#include "MSP/Basic_Types.h"
#include "MSP/HIL.h"
#include "MSP/MSP430_Debug.h"

#include "MSP/Cycles.h"
#include "MSP/Globals.h"

// #defines. ------------------------------------------------------------------

// BIT MASKS FOR ADDRESSING MODES AND DOUBLE AND SINGLE OPERAND INSTRUCTION FORMATS (As/Ad):
#define AM_RegMode      0x0000   //Register Mode Rn
#define AM_IdxMode      0x0001   //Indexed Mode X(Rn)
#define AM_IdtMode      0x0002   //Indirect Register Mode @Rn
#define AM_IdAMode      0x0003   //Indirect Autoincrement Register Mode @Rn+
#define IM_SrcMask      0x0f00   //Isolation of source register
#define IM_DstMask      0x000f   //Isolation of destination register
#define IM_AsMask       0x0030   //Isolation of source addresss mode
#define IM_AdMask       0x0080   //Isolation of destination addresss mode
#define IM_ByteOpMask   0x0040   //Isolation of byte operation identifier
#define IF_DOpCodeMask  0xf000   //OpCode of Double Instruction Format
#define IF_SOpCodeMask  0xff80   //OpCode of Single Instruction Format
#define IF_JumpCond     0xfc00   //OpCode and condition for all jump instructions
#define IM_JumpSign     0x0200   //Isolation of offset sign bit
#define IM_JumpOffset   0x01ff   //Isolation of offset

// DOUBLE OPERAND INSTRUCTION IDENTIFICATION:
#define DOI_MOV         0x4000
#define DOI_ADD         0x5000
#define DOI_ADDC        0x6000
#define DOI_SUBC        0x7000
#define DOI_SUB         0x8000
#define DOI_CMP         0x9000
#define DOI_DADD        0xA000
#define DOI_BIT         0xB000
#define DOI_BIC         0xC000
#define DOI_BIS         0xD000
#define DOI_XOR         0xE000
#define DOI_AND         0xF000

// SINGLE OPERAND INSTRUCTION IDENTIFICATION:
#define SOI_RRC         0x1000
#define SOI_SWPB        0x1080
#define SOI_RRA         0x1100
#define SOI_SXT         0x1180
#define SOI_PUSH        0x1200
#define SOI_CALL        0x1280

// NO OPERAND INSTRUCTION IDENTIFICATION:
#define NOI_RETI        0x1300

// CONDITIONAL JUMP INSTRUCTION IDENTIFICATION:
#define CJ_JNE          0x2000
#define CJ_JEQ          0x2400
#define CJ_JNC          0x2800
#define CJ_JC           0x2C00
#define CJ_JN           0x3000
#define CJ_JGE          0x3400
#define CJ_JL           0x3800
#define CJ_JMP          0x3C00

// File local variables. ------------------------------------------------------

// CYCLES FOR DOUBLE OPERAND INSTRUCTIONS:
// The following Array contains the cycles for each DOIF. To get the cycles for a DOIF,
// place As into Bit 2+3 and Ad into Bit 0 of a Byte. Bit 1 and 4 to 7 are zero. This Byte
// has to be used as index into the following array 

static unsigned int  DOICycles[16] = {
/*Src,Dst           AdAs   Index                   Cycles[AdAs] */
/*Rn,Rn             0000   0    */                 1,
/*X(Rn),Rn          0001   1    */                 3,
/*@Rn,Rn            0010   2    */                 2,
/*@Rn+,Rn           0011   3    */                 2,
/*Rn,@Rn            ----   4    not possible*/     0,
/*X(Rn),@Rn         ----   5    not possible*/     0,  
/*@Rn,@Rn           ----   6    not possible*/     0,  
/*@Rn+,@Rn          ----   7    not possible*/     0,  
/*Rn,X(Rn)      1000   8    */                 4,
/*X(Rn),X(Rn)       1001   9    */                 6,
/*@Rn,X(Rn)         1010   A    */                 5,
/*@Rn+,X(Rn)        1011   B    */                 5,
/*Rn,@Rn+           ----   C    not possible*/     0,
/*X(Rn),@Rn+        ----   D    not possible*/     0,
/*@Rn,@Rn+          ----   E    not possible*/     0,
/*@Rn+,@Rn+         ----   F    not possible*/     0
};

// CYCLES FOR SINGLE OPERAND INSTRUCTIONS:
// The following Arrays contain the cycles for each SOIF. To get the cycles for a SOIF,
// place Ad into Bit 0 and 1 of a Byte. Bit 2 to 7 are zero. This Byte has to be used as
// index into the corresponding array 

// RRC, RRA, SWPB, SXT:
static unsigned int SOICycles_1[4] = {
/*Dst         Ad   Index       Cycles[AsAd] */
/*Rn          00   0    */     1,
/*X(Rn)       01   1    */     4,
/*@Rn         10   2    */     3,
/*@Rn+        11   3    */     3
};

// PUSH:
static unsigned int SOICycles_2[4] = {
/*Dst         Ad   Index       Cycles[AsAd] */
/*Rn          00   0    */     3,
/*X(Rn)       01   1    */     5,
/*@Rn         10   2    */     4,
/*@Rn+        11   3    */     4
};

// CALL:
static unsigned int SOICycles_3[4] = {
/*Dst         Ad   Index       Cycles[AsAd] */
/*Rn          00   0    */     4,
/*X(Rn)       01   1    */     5,
/*@Rn         10   2    */     4,
/*@Rn+        11   3    */     5
};

// CYCLES FOR NO OPERAND INSTRUCTIONS:

static unsigned int NOICycles = 5;

// CYCLES FOR CONDITONIAL JUMP INSTRUCTIONS:

static unsigned int JMPCycles = 2;

// Functions. -----------------------------------------------------------------

void DOIF(WORD Instruction)
{
  WORD SrcRegNo  = (Instruction & IM_SrcMask) >> 8;
  WORD DstRegNo  = Instruction & IM_DstMask;
  WORD As        = (Instruction & IM_AsMask) >> 4;
  WORD Ad        = (Instruction & IM_AdMask) >> 7;
  //BOOL bByteInst = Instruction & IM_ByteOpMask;

    //Store used cycles of the instruction.
    if ((SrcRegNo == CG1 && As != AM_IdxMode) || SrcRegNo == CG2) As = AM_RegMode; //constant generator can be handled as register mode
    CPUCycles += DOICycles[((Ad<<3)|As)];
/*    if (DstRegNo == PC && Ad == AM_RegMode && (As == AM_RegMode || (As == AM_IdAMode && SrcRegNo != PC)))
      CPUCycles++;
*/
    if (DstRegNo == PC && Ad == AM_RegMode && (As == AM_RegMode || (As == AM_IdAMode && SrcRegNo == PC)))
      CPUCycles++;

}

void SOIF(WORD Instruction)
{
  WORD DstRegNo  = Instruction & IM_DstMask;
  WORD Ad        = (Instruction & IM_AsMask) >> 4;
  WORD SOICycles = 0;

  switch(Instruction & IF_SOpCodeMask)
  {
    case SOI_PUSH: SOICycles = 2;
           break;
    case SOI_CALL: SOICycles = 3;
           break;
    case SOI_RRC:  SOICycles = 1;
           break;
    case SOI_RRA:  SOICycles = 1;
           break;
    case SOI_SWPB: SOICycles = 1;
           break;
    case SOI_SXT:  SOICycles = 1;
           break;
    case NOI_RETI: SOICycles = 0;
           break;
    default:
           break;
  }

  if (DstRegNo == CG2 || (DstRegNo == CG1 && Ad != AM_IdxMode)) Ad = AM_RegMode; //constant generator can be handled as register mode
  switch (SOICycles)
    {
      case 1:  CPUCycles += SOICycles_1[Ad];
               break;
      case 2:  CPUCycles += SOICycles_2[Ad];
               break;
      case 3:  CPUCycles += SOICycles_3[Ad];
               break;
      default: CPUCycles += NOICycles;
               break;
    }
}

void Jumps(WORD Instruction)
{
    CPUCycles += JMPCycles;
};

void GetCycles(WORD Instruction)
{
  switch (Instruction & 0xf000)
      {
    case 0x0000:
    case 0x1000: SOIF(Instruction);  break;   //Single Operand Instruction Format
    case 0x2000:                        //Single Conditional Instruction Jump Format
    case 0x3000: Jumps(Instruction); break;
    default:     DOIF(Instruction);           //Double Operand Instruction Format
      };
}
