/*========================================================================*\
|                                                                          |
| Execution.c                                                              |
|                                                                          |
| This file contains the execution functions, and these functions contain  |
| most of the work-arounds to the device "features"                        |
| (a.k.a., device bugs)/exception conditions.                              |
|--------------------------------------------------------------------------|
| Project:              MSP430 JTAG interfcae                              |
| Developed using:      MS Visual C++ 5.0                                  |
|--------------------------------------------------------------------------|
| Author:               Dale Wellborn                                      |
| Version:              1.1.3                                              |
| Initial Version:      20 / 07 / 02                                       |
| Last Change:                                                             |
|--------------------------------------------------------------------------|
| Version history:                                                         |
| Version: 1.1.2                                                           |
| 01/20/2003 HHOF Activating of breakpoints moved from setBreakpointsActive| 
|                 to MSP430_Breakpoint                                     |
| 03/17/2003 HHOF Check for Breakpoints active changed before address != 0 |
|                 now CPUSTOP != 0                                         |
| Version: 1.1.3                                                           |
| 04/25/2003 UPSF Change read of PC to address 0xFE to avoid destroy 0x1FE |
| 24.07.2003 UPSF readAllDeviceRegisters added update the ADC10 registers  |
|                 after a PUC/ RESET was executed                          |
| Version: 1.1.5                                                           |
| 01/21/2004 UPSF added deviceInternal union                               |
|                                                                          |
|--------------------------------------------------------------------------|
| Designed 2002 by Texas Instruments                                       |
\*========================================================================*/

/*------------------------------------------------------------------------*\
| Remarks:                                                                 |
|                                                                          |
\*------------------------------------------------------------------------*/

// #includes. -----------------------------------------------------------------

#include "MSP/Basic_Types.h"
#include "MSP/HIL.h"
#include "MSP/MSP430_Debug.h"

#include "MSP/Execution.h"
#include "MSP/Globals.h"
#include "MSP/JTAG.h"
#include "MSP/JTAG_defs.h"
#include "MSP/Memory.h"

// #defines. ------------------------------------------------------------------

#define OP_BIC_W 0xC0B2			// BIC #wwww,&aaaa
#define OP_BIC_B 0xC0F2			// BIC.B #bb,&aaaa
#define OP_BIS_W 0xD0B2			// BIS #wwww,&aaaa
#define OP_BIS_B 0xD0F2			// BIS.B #bb,&aaaa

//#define bADC10DTC0       0x0048 // ADC10DTC0 Address
#define ADC10Fetch       0x0001 // Mask for ADC10DTC0 Fetch bit

//#define wADC10CTL0       0x01b0 // ADC10CTL0 Address
#define ADC10CTL0_ENC    0x0002 // Mask for ADC10CTL0 ENC bit
#define ADC10CTL0_IFG    0x0004 // Mask for ADC10CTL0 IFG bit

//#define wADC10CTL1       0x01b2 // ADC10CTL1 Address
#define ADC10CTL1_CONSEQ 0x0006 // Mask for ADC10CTL1 CONSEQ bits

// (Local) Function prototypes. -----------------------------------------------

static STATUS_T clkTclkAndCheckDTC(void);
static STATUS_T asmOpImAddr(WORD op, WORD value, WORD addr, WORD* readValue);

// Functions. -----------------------------------------------------------------

STATUS_T readAllDeviceRegisters(LONG* registers, BOOL updateSavedReg)
{
	// Read the PC (R0).
	registers[PC] = readMab();

	if (holdWatchdog() != STATUS_OK) // Preserve and hold the watchdog.
	{
		return (STATUS_ERROR);
	}

	if (deviceHasJTAGBug() && (registers[PC] >= (LONG)deviceRamStart) && (registers[PC] <= deviceRamEnd()))
	{
		setPC(ROM_ADDR); // Force the MAB to ROM_ADDR to minimize possibility of RAM corruption.
	}

	// Read the Status register (R2).
	if (cycleStart() != STATUS_OK)
	{
		return (STATUS_ERROR);
	}
//	cycle(0x4282); // MOV R2, &01FE
//	cycle(0x01fe); // This address suggested by design. -> destroys 0x1FE
	cycle(0x4282); // MOV R2, &0FE
	cycle(0x0fe); // This address suggested by design.
	HIL_TCLK(POS_EDGE);
	HIL_TCLK(0);
	HIL_JTAG_IR_(IR_DATA_CAPTURE); // Capture the output of the MOV instruction (i.e., the SR).
	registers[SR] = HIL_JTAG_DR16(0);
	cntrlSigHighByte(CNTRL_SIG_TCE1); // Restore JTAG control of read/write.
	HIL_TCLK(1);
	// Since the PC has already been read, there is no need to restore/adjust it.

	registers[R3] = 0; // Zero the Constant register (R3).
	// Read the General Purpose registers.
	HIL_JTAG_IR_(IR_CAPTURE_CPU_REG);
	HIL_JTAG_DR16(SP); // Prepare to read register SP.
	{
		int i;
		int reg[] = { SP, R4, R5, R6, R7, R8, R9, R10, R11, R12, R13, R14, R15, 0 }; // Note dummy last value.
		int sizereg = sizeof(reg) / sizeof(*reg) - 1; // Note size excludes dummy value.

		for (i = 0; i < sizereg; i++) // Read registers SP, and R4 through R15.
		{
			registers[reg[i]] = HIL_JTAG_DR16((WORD) reg[i + 1]);
		}
	}

// ****************
// added 24.07.2003 UPSF   required to update the ADC10 registers after a PUC/ RESET was executed
    if (deviceHasDTCBug() && updateSavedReg)
    {
        // Configure the DTC so that it transfers after the present CPU instruction is complete (ADC10FETCH).
        // Save and clear ADC10CTL0 and ADC10CTL1 to switch off DTC (Note: Order matters!).

        WORD tempAdc10ctl0;

        if (asmOpImAddr(OP_BIS_B, ADC10Fetch,       bADC10DTC0, &savedAdc10dtc0) != STATUS_OK) 
          return (STATUS_ERROR);
        if (asmOpImAddr(OP_BIC_W, ADC10CTL0_ENC,    wADC10CTL0, &tempAdc10ctl0 ) != STATUS_OK) 
          return (STATUS_ERROR);
        if (asmOpImAddr(OP_BIC_W, ADC10CTL1_CONSEQ, wADC10CTL1, &savedAdc10ctl1) != STATUS_OK) 
          return (STATUS_ERROR);

        // Read final value of ADC10CTL0 (Interrupt flag might be set now)
        if (asmOpImAddr(OP_BIC_W, ADC10CTL0_ENC,    wADC10CTL0, &savedAdc10ctl0) != STATUS_OK) 
          return (STATUS_ERROR);

        savedAdc10ctl0 = (tempAdc10ctl0 & ~ADC10CTL0_IFG) | (savedAdc10ctl0 & ADC10CTL0_IFG);
    }
// ****************


	return (isInstrLoad()); // (As an internal check) Must exit on an instruction load boundary.
}

STATUS_T writeAllDeviceRegisters(LONG* registers) // Including watchdog.
{
	int i;
	// Write all registers except Constant Generator (R3). The Status Register (R2) is written second to last.
	// The PC (R0) *must* be written last, and nothing must corrupt the address and data busses prior to releasing the CPU.
	// The watchdog is also restored prior to restoring the PC. Ensure that the watchdog can be restored even if the CPU
	// is Off (since the Status Register is restored prior to the restoration of the watchdog).
	int Registers[] = { R15, R14, R13, R12, R11, R10, R9, R8, R7, R6, R5, R4, SP, SR, PC };

	for (i = 0; i < sizeof(Registers) / sizeof(*Registers); i++)
	{
		if (Registers[i] == PC) // Restore the watchdog just prior to restoring the PC.
		{
			if (writeWord(WDT, 1, &wWDTCTL) != STATUS_OK)
			{
				return (STATUS_ERROR); // Could not restore watchdog!
			}
			//wWDTCTL = 0; // mark that WDTCTL should be stored the next time

            // Restore PC together with ADC10CTL0 and ADC10CTL1 if required
            if (deviceHasDTCBug())
            {
	            if (cycleStart() != STATUS_OK)
	            {
		            return (STATUS_ERROR);
	            }
                //HIL_TCLK(0);
                HIL_JTAG_IR_(IR_DATA_16BIT);
                HIL_JTAG_DR16(MOV_IMM_PC);		// mov #??, PC
				HIL_TCLK(POS_EDGE);
                HIL_JTAG_DR16(deviceRegs[PC]);	// stored PC value
				HIL_TCLK(0);
                // Halt CPU
                cntrlSigWord(CNTRL_SIG_TCE1 | CNTRL_SIG_READ | CNTRL_SIG_HALT_JTAG);
				HIL_TCLK(NEG_EDGE);

                cntrlSigWord(CNTRL_SIG_TCE1 | CNTRL_SIG_HALT_JTAG); // Write-Mode
                // Write saved value to ADC10DTC0
                HIL_JTAG_IR_(IR_ADDR_16BIT);
                HIL_JTAG_DR16(bADC10DTC0);
                HIL_JTAG_IR_(IR_DATA_TO_ADDR);
				HIL_TCLK(1);
                HIL_JTAG_DR16(savedAdc10dtc0);
				HIL_TCLK(0);
                // Write saved value to ADC10CTL1
                HIL_JTAG_IR_(IR_ADDR_16BIT);
                HIL_JTAG_DR16(wADC10CTL1);
                HIL_JTAG_IR_(IR_DATA_TO_ADDR);
				HIL_TCLK(1);
                HIL_JTAG_DR16(savedAdc10ctl1);
				HIL_TCLK(0);
                // Write saved value to ADC10CTL0
                HIL_JTAG_IR_(IR_ADDR_16BIT);
                HIL_JTAG_DR16(wADC10CTL0);
                HIL_JTAG_IR_(IR_DATA_TO_ADDR);
				HIL_TCLK(1);
                HIL_JTAG_DR16(savedAdc10ctl0);
				HIL_TCLK(0);
                // Release "Halt" of CPU
                cntrlSigWord(CNTRL_SIG_TCE1 | CNTRL_SIG_READ);

                //--------------------------------------------------------
                // NOTE: this should be the last clock edge 
                //       before the CPU is release!
				HIL_TCLK(1);
                // NOTE: this should be the last clock edge 
                //       before the CPU is release!
                //--------------------------------------------------------
            }
            else // Restore PC as usual
            { 
		        if (writeReg(PC) != STATUS_OK)
		        {
			        return (STATUS_ERROR); // Could not restore registers! 
		        }
            } 
        } // Register[i] == 0
        else
        {
		    if (writeReg(Registers[i]) != STATUS_OK)
		    {
			    return (STATUS_ERROR); // Could not restore registers!
		    }
        } // Register[i] != 0
	}

	return (STATUS_OK);
}

STATUS_T writeReg(int regNo)
{
	if (regNo == PC)
	{
		return (setPC((WORD) deviceRegs[PC]));
	}
	
	if (cycleStart() != STATUS_OK) // Must start on an instruction load boundary.
	{
		return (STATUS_ERROR);
	}
	cycle((WORD) (0x4030 | regNo)); // MOV #<word>, Rn
	cycle((WORD) deviceRegs[regNo]);
	// It is critical that the PC be preserved.
	cycle(JMP_D_4); // Restore the PC (jmp $-4).

	return (cycleEnd());
}

STATUS_T setBreakpointsActive(int nBps, WORD* bpState)
{
	int i;
	WORD Address;

	hardwareBPsUsed = FALSE;

	if (deviceEmulation() != EMEX_NONE)
	{
		*bpState = 0;

	 /*	for (i = 0; i < nBps; i++) // For each breakpoint.
		{
			if ((jtagCommand(J_INSTR, IR_EMEX_DATA_EXCHANGE, NULL)					  ||
				 jtagCommand(J_WORD, (WORD) (i * MX_BLKSIZE + MX_BP + MX_READ), NULL) || // Read breakpoint address.
				 jtagCommand(J_WORD, 0x0000, &Address)									) != STATUS_OK)
			{
				return (STATUS_ERROR);
			}
			
			if (Address != 0) // Configure breakpoint.
			{
				hardwareBPsUsed = TRUE;
			}
		} // For each breakpoint.
      */
		

		if ((jtagCommand(J_INSTR, IR_EMEX_DATA_EXCHANGE, NULL)					  ||
		     jtagCommand(J_WORD, (WORD) (MX_CPUSTOP + MX_READ), NULL) || // Enable breakpoint(s).
			 jtagCommand(J_WORD, 0x0000, bpState)						 ) != STATUS_OK)
		{
			return (STATUS_ERROR);
		}
		hardwareBPsUsed = (bpState != 0);
		if (deviceClockControl() != GCC_NONE)
		{
			if ((jtagCommand(J_WORD, (WORD) (MX_GCLKCTRL + MX_WRITE), NULL) || // Configure clock control.
				 jtagCommand(J_WORD, (WORD) clkCntrl, NULL)					  ) != STATUS_OK)
			{
				return (STATUS_ERROR);
			}
		}
		if ((jtagCommand(J_WORD, (WORD) (MX_GENCNTRL + MX_WRITE), NULL) || // Enable EMEX control of CPU.
			 jtagCommand(J_WORD, (WORD) (genCntrl | CLEAR_STOP | EEM_EN), NULL)) != STATUS_OK)
		{
			return (STATUS_ERROR);
		}
	} // if (hasEmex_E00[chipDeriv_E00])
	else
	{
		WORD ir_bp_capture[] = { IR_BP1_CAPTURE, IR_BP2_CAPTURE };

		*bpState = 0x0040; // Break on memory address bus during instruction fetch.

		for (i = 0; i < nBps; i++) // For each breakpoint.
		{
			if ((jtagCommand(J_INSTR, ir_bp_capture[i], NULL) ||
				 jtagCommand(J_WORD, 0x0000, &Address)          ) != STATUS_OK)
			{
				return (STATUS_ERROR);
			}

			if (Address != 0) // Configure breakpoint.
			{
				*bpState |= (1 << i);

				hardwareBPsUsed = TRUE;
			}
		} // For each breakpoint.

		if ((jtagCommand(J_INSTR, IR_BP_CNTL_16BIT, NULL) ||
			 jtagCommand(J_WORD, *bpState, NULL)          ) != STATUS_OK)
		{
			return (STATUS_ERROR);
		}
	}

  return (STATUS_OK);
}

STATUS_T releaseCPU(void) // Release the CPU from Jtag control (but retain Jtag control signals).
{
	if (isInstrLoad() != STATUS_OK) // Must be on an instruction load boundary.
	{
		return (STATUS_ERROR);
	}
	HIL_JTAG_IR_(IR_ADDR_CAPTURE); // Release data and memory busses from Jtag control.
	HIL_JTAG_IR_(IR_CNTRL_SIG_RELEASE);

	return (STATUS_OK);
}

#define MAX_TCE1 10

// Synchronize the Jtag with the CPU (on an instruction load boundary). TCLK becomes the CPU clock, and is input via TDI.
// Note: Since the function exits with the CPU on an instruction load boundary, READ will be set (in addition to BYTE
//       being cleared) as the CPU prepares to read the next 16-bit instruction.
// Note: It is known that when this function is called the CPU is not OFF, and that no interrupts are possible *** OR ***
//       it is known that the device will be reset immediately after this function returns.
// Note: The state of the machine (specifically, the PC) does not have to be maintained.
// It's possible that the JTAG and CPU are already synchronized when this function is called. This would be the case
//  if the CPU was stopped as a result of a breakpoint or single step. In that case, don't (re-) synchronize, as this
//  could incorrectly interprete a pending (but suspended) DTC transfer.
// *** The low byte of the cntrl_sig register contains garbage until the JTAG and CPU are synchronized. ***
// *** It's required to initialize the low byte of the cntrl-sig register after the JTAG and CPU synchronize (TCE). ***
// *** There are three events which initiate a JTAG/CPU synchronization: TCE, TCE1, and breakpoint/single-step. ***
// *** Note that JTAG instructions not supported by the device will be decoded as BYPASS, and these will casue the
//     the CPU to begin execution.

STATUS_T syncCPU(void)
{
  int i;
  WORD cntrlSig;

  HIL_JTAG_IR_(IR_CNTRL_SIG_CAPTURE);
  cntrlSig = (WORD) HIL_JTAG_DR16(0);

  HIL_TCLK(1);

  if (!(cntrlSig & CNTRL_SIG_TCE))
  {
	  // If the JTAG and CPU are not already synchronized ...
	  // Initiate Jtag and CPU synchronization. Read/Write is under CPU control. Source TCLK via TDI.
	  // Do not effect bits used by DTC (CPU_HALT, MCLKON).
	  cntrlSigHighByte(CNTRL_SIG_TCE1 | CNTRL_SIG_CPU);
//	  HIL_JTAG_IR_(IR_CNTRL_SIG_CAPTURE);
//	  cntrlSig = (WORD) HIL_JTAG_DR16(0);
	  // A bug in F43x and F44x silicon requires that Jtag synchronization be forced when the CPU is Off.
	  // Since the JTAG and CPU are not yet synchronized, the CPUOff bit in the lower byte of the cntrlSig
	  // register is not valid.

//      if (deviceHasForceSyncBug() /* && (cntrlSig & CNTRL_SIG_CPU_OFF) */ )
//	  {
//		  HIL_JTAG_IR_(IR_EMEX_DATA_EXCHANGE);
//		  HIL_JTAG_DR16(MX_GCLKCTRL);
//		  HIL_JTAG_DR16(FORCE_SYN); // Set Force Jtag Synchronization bit in Emex General Clock Control register.
//		  HIL_JTAG_DR16(MX_GCLKCTRL);
//		  HIL_JTAG_DR16(0); // Reset Force Jtag Synchronization bit in Emex General Clock Control register.
//      } 

    if (deviceHasForceSyncBug() /* && (cntrlSig & CNTRL_SIG_CPU_OFF) */ )
    {
       WORD GClkCtrl;
          
       HIL_JTAG_IR_(IR_EMEX_DATA_EXCHANGE);
       HIL_JTAG_DR16(MX_GCLKCTRL + MX_READ);
       GClkCtrl = (WORD) HIL_JTAG_DR16(0) ;
       HIL_JTAG_DR16(MX_GCLKCTRL + MX_WRITE);
       // Set Force Jtag Synchronization bit in Emex General Clock Control register.
       HIL_JTAG_DR16((WORD)(GClkCtrl | FORCE_SYN));
       HIL_JTAG_DR16(MX_GCLKCTRL + MX_WRITE);
       // Reset Force Jtag Synchronization bit in Emex General Clock Control register.
       HIL_JTAG_DR16((WORD)(GClkCtrl & ~FORCE_SYN));
    }

      
      // TCE eventually set indicates synchronization (and clocking via TCLK).
	  for (HIL_JTAG_IR_(IR_CNTRL_SIG_CAPTURE), i = 0;
	  (!((cntrlSig = (WORD) HIL_JTAG_DR16(0)) & CNTRL_SIG_TCE)) && (i < MAX_TCE1); i++)
	  {
		  HIL_DelayMSec(1);
	  }
	  if (i == MAX_TCE1)
	  {
		  return (STATUS_ERROR); // Synchronization failed!
	  }
  } // If the JTAG and CPU are not already synchronized ...

  if (cntrlSig & CNTRL_SIG_CPU_HALT) // Disable the DTC if active.
  {
	  HIL_TCLK(0);
	  // Clear HALT. Read/Write is under JTAG control. As a precaution, disable interrupts.	  
	  cntrlSigWord(CNTRL_SIG_READ | CNTRL_SIG_TCE1);
	  HIL_TCLK(1);
  }
  else
  {  
	  // Read/Write is under JTAG control. As a precaution, disable interrupts.
	  cntrlSigWord(CNTRL_SIG_READ | CNTRL_SIG_TCE1);
  }

  // Advance to an instruction load boundary. Normal the device will be on an instruction load boundary
  // following the synchronization. However, a bug in the F413 can stop the CPU off of an instruction
  // load boundary. In that case, manually advance the CPU to an instruction load boundary. Unfortunately,
  // at this time the device type can be unknown. Consequently, advance to an instruction load boundary
  // for all devices.
  if (instrLoad() != STATUS_OK)
  {
	  return (STATUS_ERROR); // Synchronization failed!
  }

  // Execute a dummy instruction (MOV R3,R3) to work around a problem in the F12x that can sometimes cause
  // the first TCLK after synchronization to be lost.
  // Note: It's critical that the dummy instruction require only a single cycle.
  // Note: Since the state of the PC is not critical at this time, there is no need to restore/adjust it.
  // Note: It is known that at this time that interrupts are not possible (GIE is clear).
  cycle(0x4303);
  //cycle(0x4303);
  //cycle(0x4303);

  return (isInstrLoad()); // (As in internal check) Must exit on an instruction load boundary.
}

// Synchronize the Jtag with the CPU (on an instruction load boundary). TCLK becomes the CPU clock, and is input via TDI.
// Note: Since the function exits with the CPU on an instruction load boundary, READ will be set (in addition to BYTE
//       being cleared) as the CPU prepares to read the next 16-bit instruction.
// Emex logic (if present) is disabled.
// Interrupts are disabled.
// If the CPU was OFF, it is turned ON, and the change is noted.
// If an interrupt is detected, it is processed. This also turns the CPU ON.
// The register contents are read within GO and SINGLE STEP. The CPUOff bit read by these functions will be set if the
//   bit was cleared within this function.
// It's possible that the JTAG and CPU are already synchronized when this function is called. This would be the case
//  if the CPU was stopped as a result of a breakpoint or single step. In that case, don't (re-) synchronize, as this
//  could incorrectly interprete a pending (but suspended) DTC transfer.
// *** The low byte of the cntrl_sig register contains garbage until the JTAG and CPU are synchronized. ***
// *** It's required to initialize the low byte of the cntrl-sig register after the JTAG and CPU synchronize (TCE). ***
// *** There are three events which initiate a JTAG/CPU synchronization: TCE, TCE1, and breakpoint/single-step. ***
// *** Note that JTAG instructions not supported by the device will be decoded as BYPASS, and these will casue the
//     the CPU to begin execution.

//------------------------------------------------------------------------------
// New synchronization (after debug) routine by Volker Rzehak
//------------------------------------------------------------------------------

STATUS_T syncCPUAfterDebug(WORD* procState, WORD* pc, BOOL* clearedCpuOff)
{
  int i;
  WORD cntrlSig;
//  BOOL extraCycle = FALSE;

  HIL_TCLK(1);

  HIL_JTAG_IR_(IR_CNTRL_SIG_CAPTURE);
  cntrlSig = (WORD) HIL_JTAG_DR16(0);

  if (!(cntrlSig & CNTRL_SIG_TCE))
  {
    // If the JTAG and CPU are not already synchronized ...
    // Initiate Jtag and CPU synchronization. Read/Write is under CPU control. Source TCLK via TDI.
    // Do not effect bits used by DTC (CPU_HALT, MCLKON).
    cntrlSigHighByte(CNTRL_SIG_TCE1 | CNTRL_SIG_CPU);
//    HIL_JTAG_IR_(IR_CNTRL_SIG_CAPTURE);
//    cntrlSig = (WORD) HIL_JTAG_DR16(0);
    // A bug in first F43x and F44x silicon requires that Jtag synchronization be forced /* when the CPU is Off */.
	// Since the JTAG and CPU are not yet synchronized, the CPUOFF bit in the lower byte of the cntrlSig
	// register is not valid.
    if (deviceHasForceSyncBug() /* && (cntrlSig & CNTRL_SIG_CPU_OFF) */ )
    {
       WORD GClkCtrl;
          
       HIL_JTAG_IR_(IR_EMEX_DATA_EXCHANGE);
       HIL_JTAG_DR16(MX_GCLKCTRL + MX_READ);
       GClkCtrl = (WORD) HIL_JTAG_DR16(0) ;
       HIL_JTAG_DR16(MX_GCLKCTRL + MX_WRITE);
       // Set Force Jtag Synchronization bit in Emex General Clock Control register.
       HIL_JTAG_DR16((WORD)(GClkCtrl | FORCE_SYN));
       HIL_JTAG_DR16(MX_GCLKCTRL + MX_WRITE);
       // Reset Force Jtag Synchronization bit in Emex General Clock Control register.
       HIL_JTAG_DR16((WORD)(GClkCtrl & ~FORCE_SYN));
    }
    
    // TCE eventually set indicates synchronization (and clocking via TCLK).
    for (HIL_JTAG_IR_(IR_CNTRL_SIG_CAPTURE), i = 0;
         (!((cntrlSig = (WORD) HIL_JTAG_DR16(0)) & CNTRL_SIG_TCE)) && (i < MAX_TCE1); 
         i++)
    {
      HIL_DelayMSec(1);
    }
    
    if (i == MAX_TCE1)
    {
      return (STATUS_ERROR); // Synchronization failed!
    }
  } // If the JTAG and CPU are not already synchronized ...
  
  if (cntrlSig & CNTRL_SIG_CPU_HALT) // Disable the DTC if active.
  {
	  HIL_TCLK(0);
	  // Clear HALT. Read/Write is under CPU control. As a precaution, disable interrupts.	  
	  cntrlSigWord(CNTRL_SIG_TCE1 | CNTRL_SIG_CPU);
	  HIL_TCLK(1);
  }
  else
  {  
	  // Read/Write is under CPU control. As a precaution, disable interrupts.
	  cntrlSigWord(CNTRL_SIG_TCE1 | CNTRL_SIG_CPU);
  }

  if (deviceHasForceInstrLoadBug())
  {
	  // A bug in the F413 can stop the CPU off of an instruction load boundary after synchronization.
	  // In that case, manually advance the CPU to an instruction load boundary.
	  if (instrLoad() != STATUS_OK)
	  {
		  return (STATUS_ERROR); // Synchronization failed!
	  }
  }
  else
  {
	  HIL_JTAG_IR_(IR_CNTRL_SIG_CAPTURE);
	  cntrlSig = (WORD) HIL_JTAG_DR16(0);
	  if (!(cntrlSig & CNTRL_SIG_INSTRLOAD)) // (As an internal check) Must be on an instruction load boundary.
	  {
		  return (STATUS_ERROR);
	  }
  }

  //-----------------------------------------------------------------------  
  // Read PC now!!! There must have been no clocking so far (other than to have cleared a set DTC)!
  *pc = readMab();

  //-----------------------------------------------------------------------  
  if (deviceEmulation() != EMEX_NONE) 
  {
    // Disable Emex in order to control CPU via JTAG.
    if (hardwareBPsUsed)
    {
      // ProcState is required within GO. However, Stopping Emex clears ProcState. Consequently, return
      // ProcState to the caller before stopping Emex.
      if ((jtagCommand(J_INSTR, IR_EMEX_READ_CONTROL, NULL) || // Was a breakpoint hit?
           jtagCommand(J_WORD, 0x0000, procState)            ) != STATUS_OK)
      {
        return (STATUS_ERROR);
      }
    }

    // Disable Emex. A bug requires that the Emex be cleared and then disabled.
    if ((jtagCommand(J_INSTR, IR_EMEX_WRITE_CONTROL, NULL)                  ||
         jtagCommand(J_WORD, (WORD) (genCntrl | CLEAR_STOP | EEM_EN), NULL) ||
         jtagCommand(J_WORD, (WORD) (genCntrl & ~EEM_EN), NULL)               ) != STATUS_OK)
    {
      return (STATUS_ERROR);
    }
    // EEM disabled!
  } 
    
  //-----------------------------------------------------------------------  
  // Execute a dummy instruction (BIS #0,R4) to work around a problem in the F12x that can sometimes cause
  // the first TCLK after synchronization to be lost. If the TCLK is lost, try the cycle again.
  // Note: It is critical that the dummy instruction require exactly two cycles to execute. The version of BIS #
  // used does not use the constant generator, and so it requires two cycles.
  // The dummy instruction also provides a needed extra clock when the device is stopped by the Emex module and it is OFF.
  // The dummy instruction also possibly initiates processing of a pending interrupt.

#define BIS_IMM_0_R4 0xd034

  // cycle(BIS_IMM_0_R4);
  HIL_JTAG_IR_(IR_DATA_16BIT);
  HIL_JTAG_DR16(BIS_IMM_0_R4);
  if (clkTclkAndCheckDTC() != STATUS_OK) return (STATUS_ERROR);

  HIL_JTAG_IR_(IR_CNTRL_SIG_CAPTURE);
  if (HIL_JTAG_DR16(0) & CNTRL_SIG_INSTRLOAD) // Still on an instruction load boundary?
  {
    // cycle(BIS_IMM_0_R4);
    HIL_JTAG_IR_(IR_DATA_16BIT);
    HIL_JTAG_DR16(BIS_IMM_0_R4);
    if (clkTclkAndCheckDTC() != STATUS_OK) return (STATUS_ERROR);
//    extraCycle = TRUE;
  }
  // cycle(0x0000);
  HIL_JTAG_IR_(IR_DATA_16BIT);
  HIL_JTAG_DR16(0x0000);
  if (clkTclkAndCheckDTC() != STATUS_OK) return (STATUS_ERROR);

  // Advance to an instruction load boundary if an interrupt is detected. 
  // The CPUOff bit will be cleared.
  // Basically, if there is an interrupt pending, the above dummy instruction 
  // will have initialted its processing, and the CPU will not be on an 
  // instruction load boundary following the dummy instruction.
  for (HIL_JTAG_IR_(IR_CNTRL_SIG_CAPTURE), i = 0;
  (!((cntrlSig = (WORD) HIL_JTAG_DR16(0)) & CNTRL_SIG_INSTRLOAD)) && (i < MAX_TCE1); i++)
  {
 	// HIL_TCLK(POS_EDGE);
    if (clkTclkAndCheckDTC() != STATUS_OK) return (STATUS_ERROR);
  }
  if (i == MAX_TCE1)
  {
      return (STATUS_ERROR); // Could not step to instruction load boudary.
  }

  if (i == 0) 
  { // An interrupt was not detected.
  
    if (cntrlSig & CNTRL_SIG_CPU_OFF)
    {
      // The CPU is OFF. Clear CPUOff bit.
      //cycle(0xc032); // BIC #CPUOFF,R2
      HIL_JTAG_IR_(IR_DATA_16BIT);
      HIL_JTAG_DR16(0xC032);
      if (clkTclkAndCheckDTC() != STATUS_OK) return (STATUS_ERROR);

      //cycle(0x0010);
      HIL_JTAG_IR_(IR_DATA_16BIT);
      HIL_JTAG_DR16(0x0010);
      if (clkTclkAndCheckDTC() != STATUS_OK) return (STATUS_ERROR);
    }
    *clearedCpuOff = cntrlSig & CNTRL_SIG_CPU_OFF; // Preserve the CPUOff bit.
  }
  else
  { // An interrupt was detected and processed. 
    // The CPUOff bit is cleared.
    // The PC needs to be read again.
    *pc= readMab();
    *clearedCpuOff = FALSE;
  }

  if (deviceHasDTCBug())
  {
    // Configure the DTC so that it transfers after the present CPU instruction is complete (ADC10FETCH).
	// Save and clear ADC10CTL0 and ADC10CTL1 to switch off DTC (Note: Order matters!).

    WORD tempAdc10ctl0;

    if (asmOpImAddr(OP_BIS_B, ADC10Fetch,       bADC10DTC0, &savedAdc10dtc0) != STATUS_OK) 
      return (STATUS_ERROR);
    if (asmOpImAddr(OP_BIC_W, ADC10CTL0_ENC,    wADC10CTL0, &tempAdc10ctl0 ) != STATUS_OK) 
      return (STATUS_ERROR);
    if (asmOpImAddr(OP_BIC_W, ADC10CTL1_CONSEQ, wADC10CTL1, &savedAdc10ctl1) != STATUS_OK) 
      return (STATUS_ERROR);

    // Read final value of ADC10CTL0 (Interrupt flag might be set now)
    if (asmOpImAddr(OP_BIC_W, ADC10CTL0_ENC,    wADC10CTL0, &savedAdc10ctl0) != STATUS_OK) 
      return (STATUS_ERROR);

    savedAdc10ctl0 = (tempAdc10ctl0 & ~ADC10CTL0_IFG) | (savedAdc10ctl0 & ADC10CTL0_IFG);
  }

  // Regain control of the CPU. Read/Write will be set, and source TCLK via TDI.
  cntrlSigHighByte(CNTRL_SIG_TCE1); 

  return (isInstrLoad()); // (As an internal check) Must exit on an instruction load boundary.
}

static STATUS_T clkTclkAndCheckDTC(void)
{
#define MAX_DTC_CYCLE 10 

  WORD cntrlSig;
  int dtc_cycle_cnt = 0;

  if (!deviceHasDTCBug())
  {
    // Device without DTC bug
 	HIL_TCLK(POS_EDGE);
  }
  else
  { // Device with DTC bug
    do 
    {
      HIL_TCLK(0);

      HIL_JTAG_IR_(IR_CNTRL_SIG_CAPTURE);
      cntrlSig = (WORD) HIL_JTAG_DR16(0);
      if ((dtc_cycle_cnt > 0) && 
          ((cntrlSig & CNTRL_SIG_CPU_HALT) == 0))
      {
        // DTC cycle completed, take over control again...
        cntrlSigWord(CNTRL_SIG_TCE1 | CNTRL_SIG_CPU);
      }
      if ((dtc_cycle_cnt == 0) &&
          ((cntrlSig & CNTRL_SIG_CPU_HALT) == CNTRL_SIG_CPU_HALT))
      {
        // DTC cycle requested, grant it...
        cntrlSigWord(CNTRL_SIG_CPU_HALT | CNTRL_SIG_TCE1 | CNTRL_SIG_CPU);
        dtc_cycle_cnt++;
      }  
      HIL_TCLK(1);
    }  
    while ((dtc_cycle_cnt < MAX_DTC_CYCLE) &&
           ((cntrlSig & CNTRL_SIG_CPU_HALT) == CNTRL_SIG_CPU_HALT));
  } // Device with DTC bug
           
  if (dtc_cycle_cnt < MAX_DTC_CYCLE)
    return (STATUS_OK);
  else
    return (STATUS_ERROR);
}

static STATUS_T asmOpImAddr(WORD op, WORD value, WORD addr, WORD* readValue)
{
  HIL_JTAG_IR_(IR_DATA_16BIT);
  HIL_JTAG_DR16(op);				// op #..., &...
  if (clkTclkAndCheckDTC() != STATUS_OK) return (STATUS_ERROR);
        
  HIL_JTAG_IR_(IR_DATA_16BIT);
  HIL_JTAG_DR16(value);			// Value
  if (clkTclkAndCheckDTC() != STATUS_OK) return (STATUS_ERROR);
        
  HIL_JTAG_IR_(IR_DATA_16BIT);
  HIL_JTAG_DR16(addr);				// Address
  if (clkTclkAndCheckDTC() != STATUS_OK) return (STATUS_ERROR);
        
  if (readValue != NULL)
  {
    HIL_TCLK(0);
    HIL_JTAG_IR_(IR_DATA_CAPTURE);
    *readValue = (WORD) HIL_JTAG_DR16(0);	// Read original register contents
  }
  if (clkTclkAndCheckDTC() != STATUS_OK) return (STATUS_ERROR);

  // Write value with cleared bits
  if (clkTclkAndCheckDTC() != STATUS_OK) return (STATUS_ERROR);

  // Must be on instruction boundary now.
  return(isInstrLoad());
} // asmOpImAddr
