/**
 *  Basic definitions for threads, channels, and alt structures
 * \file
 *  $Author: bpisupat $
 *  $Date: 2007/06/05 16:01:43 $
 *  $Revision: 1.1.1.1 $
 *
 *  This code owes a lot to Russ Cox plan9port
*/

#ifndef THREAD_H
#define THREAD_H

typedef struct Thread Thread;
typedef struct Alt Alt;
typedef struct _Altarray _Altarray;
typedef struct Channel Channel;

typedef enum {
  Init        /** Initial thread state       */,
  Dead        /** Dead Thread State          */,
  Ready       /** Ready Thread               */,
  Running     /** Current Running Thread     */,
  Rendezvous  /** Thread Queued on Alt Array */
} State;

typedef enum {
  CHANEND    /** End Marker for Blocking Alt Array     */,
  CHANSND    /** Send on a Channel                     */,   
  CHANRCV    /** Receive from a Channel                */, 
  CHANNOP    /** Unused Alt Entry                      */,  
  CHANNOBLK  /** End Marker for non-Blocking Alt Array */
} Chanop;

struct Channel {
  unsigned    elemsize   /** size of channel element     */;
  Alt         *aq        /** queue of alt structures     */;
  unsigned     nbuf      /** number in buffer (elements) */;
  unsigned     off       /** offset in buffer (bytes)    */;
  unsigned     bufsize   /** buffer size in elements     */;
  char         retval[4]  /** the rtn value on recv       */;
  char         buf[]     /** the buffer                  */;

};

struct Alt {
  Channel    *c          /** channel for this alternative */;
  void       *v          /** send/recv value              */;
  Chanop     op          /** operation                    */;
  Thread     *thread     /** pointer to thread            */;
  Alt        *next       /** next pointer                 */;
};

struct Thread {
  State     state        /** state                        */;
  void     (*run)(void)  /** Thread procedure             */;
  void     *pc           /** "program counter"            */;
  void     *data         /** program data                 */;
  int      sel           /** alt selection                */;
  Alt      *xalt         /** pointer to current alt array */;
  Alt      a[2]          /** private alt array            */;
  Thread   *next;        /** next in queue                */;
  char name;         /* name of this thing */
};

extern Thread *currentThread;

int chanalt(Alt *);
int chansend(Channel *, void *);
int channbsend(Channel *, void *);
int chanrecv(Channel *, void *);
int channbrecv(Channel *, void *);

void _threadready(Thread *);
void _threadsched(void);

#define CONCAT2(arg1, arg2) arg1 ## arg2
#define CONCAT(arg1, arg2) CONCAT2(arg1, arg2)

/**
 *  Macro to place at beginning of thread code
 */


#define threadBegin()   if (currentThread->pc) goto *(currentThread->pc)


/**
 *  Macro to save state when thread returns
 */

#define threadReturn() \
  do { \
    currentThread->pc = CONCAT(&&_L,__LINE__) ;	\
    return; \
    CONCAT(_L, __LINE__) :;				\
  } while (0)

/**
 * Macro to mark end of thread code
 */

#define threadFinish() if (currentThread) currentThread->state = Dead;

/* set name of thread  */
#define setThrdName(t,tname) t.name = tname

/**
 *  Macro to mark beginning of interrupt handler 
 *
 *  Saves currentThread
 */

#define handlerBegin() \
  static Thread _t; \
  Thread *_tc; \
  _tc = currentThread; \
  currentThread = &_t; \
  if (_t.pc) goto *(_t.pc)

/**
 * Macro to return from interrupt handler
 *  
 * Restores currentThread, saves handler state
 */

#define handlerReturn() \
  do {\
    _t.pc = CONCAT(&&_L, __LINE__ ) ;		\
    currentThread = _tc; \
    return; CONCAT(_L, __LINE__ ) :;		\
  } while (0)

/**
 * Macro to exit from interrupt handler
 *
 * Restores current thread
 */

#define handlerFinish() \
  do { currentThread = _tc;			\
    _t.pc = 0; } while (0)

/**
 * initializer for a synchonization (unbuffered) channel
 *
 * \param type  element type
 */

#define CHANNEL0(type) {sizeof(type),0}

/**
 * initializer for a byte sized channel with a single element
 * An unfortunate hack due to compiler limitations
 */

#define CHANNEL1() {1, 0, 0, 0, 1, {0}}

/**
 *  initializer for a buffered channel
 *
 * \param type  -- element type
 * \param elem   -- number of buffer elements
 */

#define CHANNELN(type,elem){sizeof(type), 0, 0, 0, elem ,{0,0,0,0 },{0, [(elem * sizeof(type)) - 1] = 0}}


/**
 *  Set the current thread priority to high (1)
 *
 * \return Old priority
 */

int splhi(void);

/**
 *  Set the current thread priority to low (0)
 *
 * \return Old priority
 */

int spllo(void);

/**
 *  Set the current thread priority 
 * \param i thread priority 
 *
 * \return Old priority
 */

int splx(int i);

#ifndef MSP430
#ifndef SPLX
#define splhi() 0
#define spllo() 0
#define splx(a) 0
#endif
#endif


#define send(c, m) if (chansend(c,m) < 0) {threadReturn();}
#define recv(c) if (chanrecv(c,(void *) (c.retval)) < 0) {threadReturn();}
//#define recv(c, m) if (chanrecv(c,m) < 0) {threadReturn();}
#define alt(a) if (chanalt(a) < 0) {threadReturn();} switch(currentThread->sel)

#endif
